/* GIMP - The GNU Image Manipulation Program
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * Figured Glass plug-in for Gimp
 * Copyright (C) 2016 Emmanuel Glynn e_glynn@yahoo.com
 *
 * Uses simplex noise from example code by Stefan Gustavson (stegu@itn.liu.se)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>
#include <math.h>

#include <string.h>


#define PLUG_IN_PROC        "plug-in-figured-glass"
#define PLUG_IN_BINARY      "Figured Glass"
#define PLUG_IN_VERSION     "January 2015, 1.0"
#define PLUG_IN_ROLE        "gimp-figured-glass"

#define PLUGIN_TITLE        "Figured Glass"

#define SCALE_WIDTH         180
#define SPIN_BUTTON_WIDTH   75

#define TILE_CACHE_SIZE 512

///********************************************** plug-in parameters ********************************************///
typedef struct
{
    gdouble   scale;
	gdouble	  disturbance;
} plugin_params;


///********************************************** gimp's plug-in info *******************************************///
///********************************************* gimp's query function ******************************************///
static void     query(void);

///********************************************* gimp's run function ********************************************///
static void     run (const gchar      *name,
                     gint              nparams,
                     const GimpParam  *param,
                     gint             *nreturn_vals,
                     GimpParam       **return_vals);

static void add_plugin_method(GimpDrawable *drawable);
static gboolean add_method_dialog(GimpDrawable *drawable);
static void add_preview_update(GimpDrawable *drawable, GimpPreview *preview_box);
static gboolean calc_undistorted_coords (gdouble lfX, gdouble lfY, gdouble lfScale, gdouble *lfdX, gdouble *lfdY);
static gdouble constrain(gdouble lfSrc, gdouble lfExtent);

///********************************************** gimp's plug-in info *******************************************///
GimpPlugInInfo PLUG_IN_INFO =
{
    NULL,         /// init() optional
    NULL,         /// quit() optional
    query,        /// called when plug-in starts and is changed
    run           /// runs plug-in
};


///******************************************set up default parameters ******************************************///
static plugin_params rbparams =
{
    5.0,
	0.5
};


static gint32       img_x1, img_y1, img_x2, img_y2, img_bpp, img_width, img_height, img_has_alpha;
static gdouble		lfScaledim;
static guint32		lSeed;

/// c macro that initialises arguments and calls PLUG_IN_INFO
MAIN()


/// query function - responsible for procedure registration and input arguments definition
/// 1. GimpParamDef stores the data so as to speed up loading. only updates if the plug-in is
/// modified. each param has 1-Type; 2-Name; 3-String describing the param
/// 2. gimp_install_procedure declares the procedure name, some description and help strings,
/// menu path where the plug-in should sit, image types handled by the plug-in, and at the
/// end, input and output parameters number, as well as the parameters descriptors.
static void query (void)
{
    static GimpParamDef args[] =
    {
        {
            GIMP_PDB_INT32,
            "run-mode",
            "The run mode { RUN-INTERACTIVE (0), RUN-NONINTERACTIVE (1) }"
        },
        {
            GIMP_PDB_IMAGE,
            "image",
            "Input image"
        },
        {
            GIMP_PDB_DRAWABLE,
            "drawable",
            "Input drawable"
        },
		{
            GIMP_PDB_FLOAT,
            "scale",
            "Scale"
        },
		{
            GIMP_PDB_FLOAT,
            "disturbance",
            "Turbulence"
        }
    };

    gimp_install_procedure (PLUG_IN_PROC,
                            PLUGIN_TITLE,
                            "Dent image",
                            "Emmanuel Glynn",
                            "Emmanuel Glynn",
                            PLUG_IN_VERSION ,
                            "_Figured Glass...",
                            "RGB*, GRAY*",
                            GIMP_PLUGIN,
                            G_N_ELEMENTS (args), 0,
                            args, NULL);

    gimp_plugin_menu_register(PLUG_IN_PROC, "<Image>/Filters/Distorts");
}


/// run function - core of the plugin
static void run (const gchar      *name,
                 gint              nparams,
                 const GimpParam  *param,
                 gint             *nreturn_vals,
                 GimpParam       **return_vals)
{
    ///
    static GimpParam    values[1];
    GimpPDBStatusType   status;
    GimpRunMode         run_mode;
    GimpDrawable        *drawable;

    status   = GIMP_PDB_SUCCESS;
    run_mode = (GimpRunMode)param[0].data.d_int32;


    values[0].type          = GIMP_PDB_STATUS;
    values[0].data.d_status = status;

    /// set mandatory output values
    *nreturn_vals = 1;
    *return_vals  = values;

    ///  get the specified drawable
    drawable = gimp_drawable_get (param[2].data.d_drawable);

    img_bpp = gimp_drawable_bpp (drawable->drawable_id);
    img_has_alpha = gimp_drawable_has_alpha (drawable->drawable_id);

	if (!gimp_drawable_mask_intersect (drawable->drawable_id, &img_x1, &img_y1, &img_width, &img_height))
    {
        return;
    }

	img_x2 = img_x1 + img_width;
	img_y2 = img_y1 + img_height;


    lfScaledim = (gdouble)MIN(img_width, img_height);
	lSeed = g_random_int();

	gimp_tile_cache_ntiles (TILE_CACHE_SIZE);

    /// set run modes
    switch( run_mode )
    {
        case GIMP_RUN_INTERACTIVE:
            /// *** get options last values if needed *** ///
            gimp_get_data (PLUG_IN_PROC, &rbparams);

            ///gimp_context_get_foreground (&rbparams.rgb);

            /// *** run the dialog *** ///
            if( !add_method_dialog(drawable) )
                return;
            break;
        case GIMP_RUN_NONINTERACTIVE:
			if ( nparams != 5 )
            {
                status = GIMP_PDB_CALLING_ERROR;
            }
            if ( status == GIMP_PDB_SUCCESS )
            {
                rbparams.scale       = param[3].data.d_float;
				rbparams.disturbance = param[4].data.d_float;
            }
            break;
        case GIMP_RUN_WITH_LAST_VALS:
            /// get options last values if needed
            gimp_get_data (PLUG_IN_PROC, &rbparams);
            break;
        default:
            break;
    }
    if ((status == GIMP_PDB_SUCCESS) && (gimp_drawable_is_rgb(drawable->drawable_id) || gimp_drawable_is_gray(drawable->drawable_id)))
    {
        add_plugin_method (drawable);

        if (run_mode != GIMP_RUN_NONINTERACTIVE)
        {
            gimp_displays_flush ();
        }
        if (run_mode == GIMP_RUN_INTERACTIVE)
        {
            gimp_set_data(PLUG_IN_PROC, &rbparams, sizeof(plugin_params ) );
        }
    }
    else if (status == GIMP_PDB_SUCCESS)
    {
        status = GIMP_PDB_EXECUTION_ERROR;
    }

    values[0].data.d_status = status;

    gimp_drawable_detach( drawable );



    return;
}


const gint32 Grad3[12][3] =
{
    {1,1,0}, {-1,1,0}, {1,-1,0}, {-1,-1,0},
    {1,0,1}, {-1,0,1}, {1,0,-1}, {-1,0,-1},
    {0,1,1}, {0,-1,1}, {0,1,-1}, {0,-1,-1}
};

static gint32 Perms[] =
{
	151,160,137,91,90,15,131,13,201,95,96,53,194,233,7,225,140,36,103,30,69,142,
    8,99,37,240,21,10,23,190,6,148,247,120,234,75,0,26,197,62,94,252,219,203,117,
    35,11,32,57,177,33,88,237,149,56,87,174,20,125,136,171,168,68,175,74,165,71,
    134,139,48,27,166,77,146,158,231,83,111,229,122,60,211,133,230,220,105,92,41,
    55,46,245,40,244,102,143,54,65,25,63,161,1,216,80,73,209,76,132,187,208, 89,
    18,169,200,196,135,130,116,188,159,86,164,100,109,198,173,186,3,64,52,217,226,
    250,124,123,5,202,38,147,118,126,255,82,85,212,207,206,59,227,47,16,58,17,182,
    189,28,42,223,183,170,213,119,248,152,2,44,154,163,70,221,153,101,155,167,43,
    172,9,129,22,39,253,19,98,108,110,79,113,224,232,178,185,112,104,218,246,97,
    228,251,34,242,193,238,210,144,12,191,179,162,241,81,51,145,235,249,14,239,
    107,49,192,214,31,181,199,106,157,184,84,204,176,115,121,50,45,127,4,150,254,
    138,236,205,93,222,114,67,29,24,72,243,141,128,195,78,66,215,61,156,180
};



static gint32 P0[ 512 ];
static gint32 P1[ 512 ];


static gdouble Grad (const gint32* g, gdouble x, gdouble z)
{
	return g[0]*x + g[1]*z;
}

static void CalculateSimplexNoisePerms(void)
{
    gint32 i;

    for( i = 0; i < 512; i++ )
	{
		P0[ i ] = Perms[(lSeed + i) & 255];
		P1[ i ] = P0[ i ] % 12;
	}
}


static gint32 FastFloor32(const gdouble x)
{
    return x > 0 ? (gint32)x : (gint32)x - 1;
}


#define F2 0.366025403 // F2 = 0.5*(sqrt(3.0)-1.0)
#define G2 0.211324865 // G2 = (3.0-sqrt(3.0))/6.0


static gdouble SimplexNoise2D (gdouble dX, gdouble dY)
{
	gdouble N0, N1, N2;

	gdouble S = (dX + dY)*F2;
	gdouble XS = dX + S;
	gdouble YS = dY + S;
	gint32 iX = FastFloor32(XS);
	gint32 iY = FastFloor32(YS);

	gdouble T = (gdouble)(iX + iY)*G2;
	gdouble X0 = iX - T;
	gdouble Y0 = iY - T;
	gdouble dX0 = dX - X0;
	gdouble dY0 = dY - Y0;


	gint32 iX1, iY1;
	if (dX0 > dY0)
	{
		iX1 = 1;
		iY1 = 0;
	}
	else
	{
		iX1 = 0;
		iY1 = 1;
	}


	gdouble X1 = dX0 - iX1 + G2;
	gdouble Y1 = dY0 - iY1 + G2;
	gdouble X2 = dX0 - 1.0 + 2.0 * G2;
	gdouble Y2 = dY0 - 1.0 + 2.0 * G2;


	gint32 II = iX & 0xff;
	gint32 JJ = iY & 0xff;

	gint32 GI0 = P1[II+P0[JJ]];
	gint32 GI1 = P1[II+iX1+P0[JJ+iY1]];
	gint32 GI2 = P1[II+1+P0[JJ+1]];

	gdouble T0 = 0.5 - dX0*dX0 - dY0*dY0;
	if (T0 < 0.0)
	{
		N0 = 0.0;
	}
	else
	{
		T0 *= T0;
		N0 = T0 * T0 * Grad(Grad3[GI0], dX0, dY0);
	}

	gdouble T1 = 0.5 - X1*X1 - Y1*Y1;
	if (T1 < 0.0)
	{
		 N1 = 0.0;
	}
	else
	{
		T1 *= T1;
		N1 = T1 * T1 * Grad(Grad3[GI1], X1, Y1);
	}

	gdouble T2 = 0.5 - X2*X2 - Y2*Y2;
	if (T2 < 0.0)
	{
		 N2 = 0.0;
	}
	else
	{
		T2 *= T2;
		N2 = T2 * T2 * Grad(Grad3[GI2], X2, Y2);
	}


	/// Interval [-1,1].
	return 70.0 * (N0 + N1 + N2);
}


static void
add_plugin_method(GimpDrawable *drawable)
{
    GimpPixelRgn        des_rgn;
	gpointer			p;
    gint32              x, y, ix, iy;
	gdouble				cx, cy;
	gdouble 			lfScale;
	guchar				pixel[4][4];
	guchar             *pixel_elems[4];
	guchar             *d, *des;
	GimpPixelFetcher   *px;


    /// ******************************** initialize PixelRgn to write output data ****************************** ///
	gimp_pixel_rgn_init (&des_rgn, drawable, img_x1, img_y1, img_width, img_height, TRUE, TRUE);


	lfScale = (rbparams.scale/100)*lfScaledim;

	gimp_progress_init ("Distorting...");

	px = gimp_pixel_fetcher_new (drawable, FALSE);
    gimp_pixel_fetcher_set_edge_mode (px, GIMP_PIXEL_FETCHER_EDGE_SMEAR);

    CalculateSimplexNoisePerms();

	for (p = gimp_pixel_rgns_register (1, &des_rgn); p != NULL; p = gimp_pixel_rgns_process (p))
	{
		des = des_rgn.data;
		for (y = des_rgn.y; y < (des_rgn.y + des_rgn.h); ++y)
		{
			d = des;
			for (x = des_rgn.x; x < (des_rgn.x + des_rgn.w); ++x)
			{
				calc_undistorted_coords ((gdouble)x, (gdouble)y, lfScale, &cx, &cy);
				{
					if (cx >= 0.0)
					{
						ix = (gint) cx;
					}
					else
					{
						ix = -((gint) -cx + 1);
					}

					if (cy >= 0.0)
					{
						iy = (gint) cy;
					}
					else
					{
						iy = -((gint) -cy + 1);
					}
					pixel_elems[0] = pixel[0];
					pixel_elems[1] = pixel[1];
					pixel_elems[2] = pixel[2];
					pixel_elems[3] = pixel[3];

					gimp_pixel_fetcher_get_pixel (px, ix    , iy    , pixel[0]);
					gimp_pixel_fetcher_get_pixel (px, ix + 1, iy    , pixel[1]);
					gimp_pixel_fetcher_get_pixel (px, ix    , iy + 1, pixel[2]);
					gimp_pixel_fetcher_get_pixel (px, ix + 1, iy + 1, pixel[3]);

					gimp_bilinear_pixels_8 (d, cx, cy, des_rgn.bpp, img_has_alpha, pixel_elems);
				}
				d += des_rgn.bpp;
			}
			des += des_rgn.rowstride;
		}
		if (y % 32 == 0)
        {
            gimp_progress_update ((gdouble) y / (gdouble) img_height);
        }
	}

    gimp_pixel_fetcher_destroy (px);

    gimp_drawable_flush (drawable);
    gimp_drawable_merge_shadow (drawable->drawable_id, TRUE);
    gimp_drawable_update (drawable->drawable_id, img_x1, img_y1, img_width, img_height);

    return;
}

static gdouble constrain(gdouble lfSrc, gdouble lfExtent)
{
	lfExtent--;

	if(lfExtent)
	{
		lfSrc = abs(lfSrc);

		do
		{
			lfSrc = abs((2 * lfExtent) - lfSrc);
		} while(lfSrc > lfExtent);
	}
	return lfSrc;
}



static gboolean calc_undistorted_coords (gdouble lfX, gdouble lfY, gdouble lfScale, gdouble *lfdX, gdouble *lfdY)
{
	gdouble				N0, N1, N2, lfScaleX, lfScaleY;
	gboolean			isinside;

	isinside = (lfScale > 0);

	if (isinside)
	{
		lfScaleX = lfX / lfScale;
		lfScaleY = lfY / lfScale;
	    N0 = SimplexNoise2D(lfScaleX  ,   lfScaleY) * rbparams.disturbance;
		N1 = SimplexNoise2D(lfScaleX + 1, lfScaleY) * rbparams.disturbance;
		N2 = SimplexNoise2D(lfScaleX, lfScaleY + 1) * rbparams.disturbance;

		*lfdX = constrain((lfX + (N1 - N0) * lfScale), img_x2);
		*lfdY = constrain((lfY + (N2 - N0) * lfScale * rbparams.disturbance), img_y2);
	}
	else
	{
		*lfdX = lfX;
		*lfdY = lfY;
	}

	return isinside;
}


gboolean add_method_dialog(GimpDrawable *drawable)
{
    gboolean    run;

    GtkWidget   *disp_box            = NULL;
    GtkWidget   *dialog              = NULL;
    GtkWidget   *dialog_vbox         = NULL;

    GtkWidget   *scale_2x3table      = NULL;
    GtkObject   *scale               = NULL;



    /// **************************************** initialize plug-in ******************************************** ///
    gimp_ui_init (PLUG_IN_BINARY, FALSE);


    dialog = gimp_dialog_new(PLUGIN_TITLE, PLUG_IN_ROLE ,
                             NULL, (GtkDialogFlags) 0,
                             gimp_standard_help_func, PLUG_IN_PROC,
                             GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                             GTK_STOCK_OK,     GTK_RESPONSE_OK,
                             NULL);

    gtk_dialog_set_alternative_button_order (GTK_DIALOG (dialog),
                                             GTK_RESPONSE_OK,
                                             GTK_RESPONSE_CANCEL,
                                             -1);

    gimp_window_set_transient (GTK_WINDOW (dialog));

    /// *************************************** create main horizontal box ************************************* ///
    dialog_vbox = gtk_vbox_new (FALSE, 12);
    /// ******************************************* set border width ******************************************* ///
    gtk_container_set_border_width (GTK_CONTAINER (dialog_vbox), 12);
    /// ************************************ add dialogue box to main box ************************************** ///
    gtk_box_pack_start (GTK_BOX (gtk_dialog_get_content_area (GTK_DIALOG (dialog))), dialog_vbox, TRUE, TRUE, 0);
    /// ***************************************** show main box ************************************************ ///
    gtk_widget_show(dialog_vbox);

    /// ***************************************** prepare preview ********************************************** ///
    disp_box = gimp_zoom_preview_new (drawable);
    ///gtk_widget_set_size_request (disp_box, 450, 450); /// change size of preview area
    /// *************************************** pack preview box into container ******************************** ///
    gtk_box_pack_start (GTK_BOX (dialog_vbox), disp_box, TRUE, TRUE, 0);

    g_signal_connect_swapped (disp_box, "invalidated", G_CALLBACK (add_preview_update), drawable);



    scale_2x3table = gtk_table_new (2, 3, FALSE);
    gtk_table_set_col_spacings (GTK_TABLE (scale_2x3table), 6);
    gtk_table_set_row_spacings (GTK_TABLE (scale_2x3table), 6);
    gtk_box_pack_start (GTK_BOX (dialog_vbox), scale_2x3table, FALSE, FALSE, 0);
    gtk_widget_show (scale_2x3table);

    scale = gimp_scale_entry_new (GTK_TABLE (scale_2x3table),
                                  0,
                                  0,
                                  ("Scale:"),
                                  SCALE_WIDTH,
                                  SPIN_BUTTON_WIDTH,
                                  rbparams.scale,
                                  0.0,
								  100.0,
                                  0.1,
                                  10,
                                  1,
                                  TRUE,
                                  0,
                                  0,
                                  NULL,
                                  NULL);
    g_signal_connect (scale, "value_changed", G_CALLBACK (gimp_double_adjustment_update), &rbparams.scale);
    g_signal_connect_swapped (scale, "value-changed", G_CALLBACK (gimp_preview_invalidate), disp_box);


	scale = gimp_scale_entry_new (GTK_TABLE (scale_2x3table),
                                  0,
                                  1,
                                  ("Turbulence:"),
                                  SCALE_WIDTH,
                                  SPIN_BUTTON_WIDTH,
                                  rbparams.disturbance,
                                  0.0,
								  1.0,
                                  0.01,
                                  0.1,
                                  2,
                                  TRUE,
                                  0,
                                  0,
                                  NULL,
                                  NULL);
    g_signal_connect (scale, "value_changed", G_CALLBACK (gimp_double_adjustment_update), &rbparams.disturbance);
    g_signal_connect_swapped (scale, "value-changed", G_CALLBACK (gimp_preview_invalidate), disp_box);

    /// ***************************************** show dialog box ********************************************** ///
    gtk_widget_show(disp_box);

    /// ******************************************** run dialog ************************************************ ///
    run = (gimp_dialog_run(GIMP_DIALOG(dialog)) == GTK_RESPONSE_OK);
    /// ********************************************* clean up ************************************************* ///
    gtk_widget_destroy(dialog);

    return run;
}


static void add_preview_update(GimpDrawable *drawable, GimpPreview *preview_box)
{
    gint32     			width, height;
    gint32     			bpp;
    gint32    			x, y, px, py;
    gdouble             ix, iy;
    gdouble				lfScale;
	guchar				pixel0[4], pixel1[4], pixel2[4], pixel3[4];
	guchar			   *pixel_elems[4];
	guchar   		   *des, *d;
	GimpPixelFetcher   *p;

	pixel_elems[0] = pixel0;
	pixel_elems[1] = pixel1;
	pixel_elems[2] = pixel2;
	pixel_elems[3] = pixel3;

    p = gimp_pixel_fetcher_new (drawable, FALSE);
    gimp_pixel_fetcher_set_edge_mode (p, GIMP_PIXEL_FETCHER_EDGE_SMEAR);

    des  = gimp_zoom_preview_get_source (GIMP_ZOOM_PREVIEW (preview_box), &width, &height, &bpp);

	lfScale = (rbparams.scale/100)*lfScaledim;

	CalculateSimplexNoisePerms();

	d = des;
	for (y = 0; y < height; y++)
    {
		for (x = 0; x < width; x++)
		{
			gimp_preview_untransform (preview_box, x, y, &px, &py);
			if(calc_undistorted_coords ((gdouble)px, (gdouble)py, lfScale, &ix, &iy))
            {
                gimp_pixel_fetcher_get_pixel (p, ix    , iy    , pixel0);
                gimp_pixel_fetcher_get_pixel (p, ix + 1, iy    , pixel1);
                gimp_pixel_fetcher_get_pixel (p, ix    , iy + 1, pixel2);
                gimp_pixel_fetcher_get_pixel (p, ix + 1, iy + 1, pixel3);


                gimp_bilinear_pixels_8 (d, ix, iy, bpp, img_has_alpha, pixel_elems);
            }

			d += bpp;
		}
	}

    gimp_pixel_fetcher_destroy (p);

    gimp_preview_draw_buffer (preview_box, des, width * bpp);

    g_free (des);
}
