/* GIMP - The GNU Image Manipulation Program
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * Pinch plug-in for Gimp
 * Copyright (C) 2016 Emmanuel Glynn e_glynn@yahoo.com
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>
#include <math.h>


#define PLUG_IN_PROC        "plug-in-pinch"
#define PLUG_IN_BINARY      "Pinch"
#define PLUG_IN_VERSION     "June 2014, 1.1"
#define PLUG_IN_ROLE        "gimp-pinch"

#define PLUGIN_TITLE        "Pinch"


#define SCALE_WIDTH         180
#define SPIN_BUTTON_WIDTH   75



///********************************************** plug-in parameters ********************************************///
typedef struct
{
    gint32   amount;
} plugin_params;


///********************************************** gimp's plug-in info *******************************************///
///********************************************* gimp's query function ******************************************///
static void     query(void);

///********************************************* gimp's run function ********************************************///
static void     run (const gchar      *name,
                     gint              nparams,
                     const GimpParam  *param,
                     gint             *nreturn_vals,
                     GimpParam       **return_vals);

static void add_plugin_method(GimpDrawable *drawable);
static gboolean add_method_dialog(GimpDrawable *drawable);
static void add_preview_update(GimpDrawable *drawable, GimpPreview *preview_box);
static gboolean calc_undistorted_coords (gdouble lfIX, gdouble lfIY, gdouble lfAmount, gdouble *lfOX, gdouble *lfOY);
static gdouble squared(const gdouble x);

///********************************************** gimp's plug-in info *******************************************///
GimpPlugInInfo PLUG_IN_INFO =
{
    NULL,         /// init() optional
    NULL,         /// quit() optional
    query,        /// called when plug-in starts and is changed
    run           /// runs plug-in
};


///******************************************set up default parameters ******************************************///
static plugin_params rbparams =
{
    50
};


static gint32       img_x1, img_z1, img_x2, img_z2, img_bpp, img_width, img_height, img_has_alpha;
static gdouble		lfCenterX, lfCenterY;
static gdouble		lfXScale, lfYScale;
static gdouble		lfRadius, lfRadMax;

/// c macro that initialises arguments and calls PLUG_IN_INFO
MAIN()


/// query function - responsible for procedure registration and input arguments definition
/// 1. GimpParamDef stores the data so as to speed up loading. only updates if the plug-in is
/// modified. each param has 1-Type; 2-Name; 3-String describing the param
/// 2. gimp_install_procedure declares the procedure name, some description and help strings,
/// menu path where the plug-in should sit, image types handled by the plug-in, and at the
/// end, input and output parameters number, as well as the parameters descriptors.
static void query (void)
{
    static GimpParamDef args[] =
    {
        {
            GIMP_PDB_INT32,
            "run-mode",
            "The run mode { RUN-INTERACTIVE (0), RUN-NONINTERACTIVE (1) }"
        },
        {
            GIMP_PDB_IMAGE,
            "image",
            "Input image"
        },
        {
            GIMP_PDB_DRAWABLE,
            "drawable",
            "Input drawable"
        },
		{
            GIMP_PDB_INT32,
            "amount",
            "Amount"
        }
    };

    gimp_install_procedure (PLUG_IN_PROC,
                            PLUGIN_TITLE,
                            "Pinch image",
                            "Emmanuel Glynn",
                            "Emmanuel Glynn",
                            PLUG_IN_VERSION ,
                            "Pinch...",
                            "RGB*, GRAY*",
                            GIMP_PLUGIN,
                            G_N_ELEMENTS (args), 0,
                            args, NULL);

    gimp_plugin_menu_register(PLUG_IN_PROC, "<Image>/Filters/Distorts");
}


/// run function - core of the plugin
static void run (const gchar      *name,
                 gint              nparams,
                 const GimpParam  *param,
                 gint             *nreturn_vals,
                 GimpParam       **return_vals)
{
    ///
    static GimpParam    values[1];
    GimpPDBStatusType   status;
    GimpRunMode         run_mode;
    GimpDrawable        *drawable;

    status   = GIMP_PDB_SUCCESS;
    run_mode = (GimpRunMode)param[0].data.d_int32;


    values[0].type          = GIMP_PDB_STATUS;
    values[0].data.d_status = status;

    /// set mandatory output values
    *nreturn_vals = 1;
    *return_vals  = values;

    ///  get the specified drawable
    drawable = gimp_drawable_get (param[2].data.d_drawable);

    img_bpp = gimp_drawable_bpp (drawable->drawable_id);
    img_has_alpha = gimp_drawable_has_alpha (drawable->drawable_id);

	if (!gimp_drawable_mask_intersect (drawable->drawable_id, &img_x1, &img_z1, &img_width, &img_height))
    {
        return;
    }

	img_x2 = img_x1 + img_width;
	img_z2 = img_z1 + img_height;

	lfCenterX = (gdouble)(img_x1 + img_x2 - 1) / 2.0;
	lfCenterY = (gdouble)(img_z1 + img_z2 - 1) / 2.0;

    gdouble xhsize, yhsize;
    xhsize = (gdouble) (img_width - 1) / 2.0;
    yhsize = (gdouble) (img_height - 1) / 2.0;

	if (xhsize > yhsize)
    {
        lfYScale = xhsize / yhsize;
        lfXScale = 1.0;
    }
    else if (yhsize > xhsize)
    {
        lfXScale = yhsize / xhsize;
        lfYScale = 1.0;
    }
    else
    {
        lfXScale = 1.0;
        lfYScale = 1.0;
    }

	lfRadius = (gdouble)MAX(xhsize, yhsize);
	lfRadMax = squared(lfRadius);

	gimp_tile_cache_ntiles (4 * drawable->ntile_cols);

    /// set run modes
    switch( run_mode )
    {
        case GIMP_RUN_INTERACTIVE:
            /// *** get options last values if needed *** ///
            gimp_get_data (PLUG_IN_PROC, &rbparams);

            /// *** run the dialog *** ///
            if( !add_method_dialog(drawable) )
                return;
            break;
        case GIMP_RUN_NONINTERACTIVE:
			if ( nparams != 4 )
            {
                status = GIMP_PDB_CALLING_ERROR;
            }
            if ( status == GIMP_PDB_SUCCESS )
            {
                rbparams.amount       = param[3].data.d_int32;
            }
            break;
        case GIMP_RUN_WITH_LAST_VALS:
            /// get options last values if needed
            gimp_get_data (PLUG_IN_PROC, &rbparams);
            break;
        default:
            break;
    }
    if ((status == GIMP_PDB_SUCCESS) && (gimp_drawable_is_rgb(drawable->drawable_id) || gimp_drawable_is_gray(drawable->drawable_id)))
    {
        add_plugin_method (drawable);

        if (run_mode != GIMP_RUN_NONINTERACTIVE)
        {
            gimp_displays_flush ();
        }
        if (run_mode == GIMP_RUN_INTERACTIVE)
        {
            gimp_set_data(PLUG_IN_PROC, &rbparams, sizeof(plugin_params ) );
        }
    }
    else if (status == GIMP_PDB_SUCCESS)
    {
        status = GIMP_PDB_EXECUTION_ERROR;
    }

    values[0].data.d_status = status;

    gimp_drawable_detach( drawable );



    return;
}


static void
add_plugin_method(GimpDrawable *drawable)
{
    GimpPixelRgn        des_rgn;
	gpointer			p;
    gint32              x, y, ix, iy;
	gdouble				cx, cy;
	gdouble 			lfAmount;
	guchar				pixel[4][4];
	guchar             *pixel_elems[4];
	guchar             *d, *des;
	GimpPixelFetcher   *px;


    /// ******************************** initialize PixelRgn to write output data ****************************** ///
	gimp_pixel_rgn_init (&des_rgn, drawable, img_x1, img_z1, img_width, img_height, TRUE, TRUE);

	lfAmount = -(gdouble)rbparams.amount / 127.0;

	gimp_progress_init ("Distorting...");

	px = gimp_pixel_fetcher_new (drawable, FALSE);
    gimp_pixel_fetcher_set_edge_mode (px, GIMP_PIXEL_FETCHER_EDGE_SMEAR);

	for (p = gimp_pixel_rgns_register (1, &des_rgn); p != NULL; p = gimp_pixel_rgns_process (p))
	{
		des = des_rgn.data;
		for (y = des_rgn.y; y < (des_rgn.y + des_rgn.h); ++y)
		{
			d = des;
			for (x = des_rgn.x; x < (des_rgn.x + des_rgn.w); ++x)
			{
				//progress++;

				calc_undistorted_coords ((gdouble)x, (gdouble)y, lfAmount, &cx, &cy);

                if (cx >= 0.0)
                {
                    ix = (gint) cx;
                }
                else
                {
                    ix = -((gint) -cx + 1);
                }

                if (cy >= 0.0)
                {
                    iy = (gint) cy;
                }
                else
                {
                    iy = -((gint) -cy + 1);
                }
                pixel_elems[0] = pixel[0];
                pixel_elems[1] = pixel[1];
                pixel_elems[2] = pixel[2];
                pixel_elems[3] = pixel[3];

                gimp_pixel_fetcher_get_pixel (px, ix    , iy    , pixel[0]);
                gimp_pixel_fetcher_get_pixel (px, ix + 1, iy    , pixel[1]);
                gimp_pixel_fetcher_get_pixel (px, ix    , iy + 1, pixel[2]);
                gimp_pixel_fetcher_get_pixel (px, ix + 1, iy + 1, pixel[3]);

                gimp_bilinear_pixels_8 (d, cx, cy, des_rgn.bpp, img_has_alpha, pixel_elems);

				d += des_rgn.bpp;
			}
			des += des_rgn.rowstride;
		}
		if (y % 16 == 0)
        {
            gimp_progress_update ((gdouble) y / (gdouble) img_height);
        }
	}

    gimp_pixel_fetcher_destroy (px);

    gimp_drawable_flush (drawable);
    gimp_drawable_merge_shadow (drawable->drawable_id, TRUE);
    gimp_drawable_update (drawable->drawable_id, img_x1, img_z1, img_width, img_height);

    return;
}

static gdouble squared(const gdouble x)
{
    return x*x;
}

static gboolean calc_undistorted_coords (gdouble ix, gdouble iy, gdouble amount, gdouble *x, gdouble *y)
{
    gdouble             dx;
    gdouble             dy;
	gboolean			isinside;
	gdouble             lfCurrentRadius, lfScale;

	dy = lfYScale*(gdouble)(iy - lfCenterY);
	dx = lfXScale*(gdouble)(ix - lfCenterX);

	lfCurrentRadius = squared(dx) + squared(dy);

	isinside = (lfCurrentRadius <= lfRadMax);

	if (isinside)
	{
		lfCurrentRadius = sqrt(lfCurrentRadius);
		lfScale = 1.0 - squared(lfCurrentRadius / lfRadius);
		lfScale = 1.0 - amount * squared(lfScale);
		*x = (dx * lfScale)/lfXScale + lfCenterX;
		*y = (dy * lfScale)/lfYScale + lfCenterY;
	}
	else
	{
		*x = ix;
		*y = iy;
	}

	return isinside;
}

static gboolean add_method_dialog(GimpDrawable *drawable)
{
    gboolean    run;

    GtkWidget   *disp_box            = NULL;
    GtkWidget   *dialog              = NULL;
    GtkWidget   *dialog_vbox         = NULL;

    GtkWidget   *scale_1x3table      = NULL;
    GtkObject   *scale               = NULL;



    /// **************************************** initialize plug-in ******************************************** ///
    gimp_ui_init (PLUG_IN_BINARY, FALSE);


    dialog = gimp_dialog_new(PLUGIN_TITLE, PLUG_IN_ROLE,
                             NULL, (GtkDialogFlags) 0,
                             gimp_standard_help_func, PLUG_IN_PROC,
                             GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                             GTK_STOCK_OK,     GTK_RESPONSE_OK,
                             NULL);

    gtk_dialog_set_alternative_button_order (GTK_DIALOG (dialog),
                                             GTK_RESPONSE_OK,
                                             GTK_RESPONSE_CANCEL,
                                             -1);

    gimp_window_set_transient (GTK_WINDOW (dialog));

    /// *************************************** create main horizontal box ************************************* ///
    dialog_vbox = gtk_vbox_new (FALSE, 12);
    /// ******************************************* set border width ******************************************* ///
    gtk_container_set_border_width (GTK_CONTAINER (dialog_vbox), 12);
    /// ************************************ add dialogue box to main box ************************************** ///
    gtk_box_pack_start (GTK_BOX (gtk_dialog_get_content_area (GTK_DIALOG (dialog))), dialog_vbox, TRUE, TRUE, 0);
    /// ***************************************** show main box ************************************************ ///
    gtk_widget_show(dialog_vbox);

    /// ***************************************** prepare preview ********************************************** ///
    disp_box = gimp_zoom_preview_new (drawable);
    ///gtk_widget_set_size_request (disp_box, 450, 450); /// change size of preview area
    /// *************************************** pack preview box into container ******************************** ///
    gtk_box_pack_start (GTK_BOX (dialog_vbox), disp_box, TRUE, TRUE, 0);

    g_signal_connect_swapped (disp_box, "invalidated", G_CALLBACK (add_preview_update), drawable);



    scale_1x3table = gtk_table_new (1, 3, FALSE);
    gtk_table_set_col_spacings (GTK_TABLE (scale_1x3table), 6);
    gtk_table_set_row_spacings (GTK_TABLE (scale_1x3table), 6);
    gtk_box_pack_start (GTK_BOX (dialog_vbox), scale_1x3table, FALSE, FALSE, 0);
    gtk_widget_show (scale_1x3table);

    scale = gimp_scale_entry_new (GTK_TABLE (scale_1x3table),
                                  0,
                                  0,
                                  ("Amount:"),
                                  SCALE_WIDTH,
                                  SPIN_BUTTON_WIDTH,
                                  rbparams.amount,
                                  -100,
                                  100,
                                  1,
                                  8,
                                  0,
                                  TRUE,
                                  0,
                                  0,
                                  NULL,
                                  NULL);
    g_signal_connect (scale, "value_changed", G_CALLBACK (gimp_int_adjustment_update), &rbparams.amount);
    g_signal_connect_swapped (scale, "value-changed", G_CALLBACK (gimp_preview_invalidate), disp_box);

    /// ***************************************** show dialog box ********************************************** ///
    gtk_widget_show(disp_box);

    /// ******************************************** run dialog ************************************************ ///
    run = (gimp_dialog_run(GIMP_DIALOG(dialog)) == GTK_RESPONSE_OK);
    /// ********************************************* clean up ************************************************* ///
    gtk_widget_destroy(dialog);

    return run;
}


static void add_preview_update(GimpDrawable *drawable, GimpPreview *preview_box)
{
    gint32     			width, height;
    gint32     			bpp;
    gint32    			x, y, px, py;
    gdouble             ix, iy;
    gdouble				lfAmount;
	guchar				pixel0[4], pixel1[4], pixel2[4], pixel3[4];
	guchar			   *pixel_elems[4];
	guchar   		   *des, *d;
	GimpPixelFetcher   *p;

	pixel_elems[0] = pixel0;
	pixel_elems[1] = pixel1;
	pixel_elems[2] = pixel2;
	pixel_elems[3] = pixel3;

    p = gimp_pixel_fetcher_new (drawable, FALSE);
    gimp_pixel_fetcher_set_edge_mode (p, GIMP_PIXEL_FETCHER_EDGE_SMEAR);

    des  = gimp_zoom_preview_get_source (GIMP_ZOOM_PREVIEW (preview_box), &width, &height, &bpp);

	lfAmount = -(gdouble)rbparams.amount / 127.0;

	d = des;
	for (y = 0; y < height; ++y)
    {
		for (x = 0; x < width; ++x)
		{
			gimp_preview_untransform (preview_box, x, y, &px, &py);
			if(calc_undistorted_coords ((gdouble)px, (gdouble)py, lfAmount, &ix, &iy))
            {
                gimp_pixel_fetcher_get_pixel (p, ix    , iy    , pixel0);
                gimp_pixel_fetcher_get_pixel (p, ix + 1, iy    , pixel1);
                gimp_pixel_fetcher_get_pixel (p, ix    , iy + 1, pixel2);
                gimp_pixel_fetcher_get_pixel (p, ix + 1, iy + 1, pixel3);


                gimp_bilinear_pixels_8 (d, ix, iy, bpp, img_has_alpha, pixel_elems);
            }

			d += bpp;
		}
	}

    gimp_pixel_fetcher_destroy (p);

    gimp_preview_draw_buffer (preview_box, des, width * bpp);

    g_free (des);
}


