/* GIMP - The GNU Image Manipulation Program
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * ZigZag plug-in for Gimp
 * Copyright (C) 2016 Emmanuel Glynn e_glynn@yahoo.com
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */



#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>
#include <math.h>

#define G_2_PI (G_PI+G_PI)


#define PLUG_IN_PROC        "plug-in-zigzag"
#define PLUG_IN_BINARY      "zigzag"
#define PLUG_IN_VERSION     "December 2014, 1.0"
#define PLUG_IN_ROLE        "gimp-zigzag"

#define PLUGIN_TITLE        "ZigZag"


#define SCALE_WIDTH         180
#define SPIN_BUTTON_WIDTH   75



///********************************************** plug-in parameters ********************************************///
typedef enum
{
    AROUND_CENTRE    = 0,
    OUT_FROM_CENTRE  = 1,
    POND_RIPPLES     = 2
} adjustment_methods;

typedef struct
{
    gint32   amount;
	gint32	 ridges;
	gint32   method;
} plugin_params;


///********************************************** gimp's plug-in info *******************************************///
///********************************************* gimp's query function ******************************************///
static void     query(void);

///********************************************* gimp's run function ********************************************///
static void     run (const gchar      *name,
                     gint              nparams,
                     const GimpParam  *param,
                     gint             *nreturn_vals,
                     GimpParam       **return_vals);

static void add_plugin_method(GimpDrawable *drawable);
static gboolean add_method_dialog(GimpDrawable *drawable);
static void add_preview_update(GimpDrawable *drawable, GimpPreview *preview_box);
static gboolean calc_undistorted_coords_pond_ripples (gdouble ix, gdouble iz, gdouble amount, gdouble ridges, gdouble *x, gdouble *z);
static gboolean calc_undistorted_coords_around_center (gdouble ix, gdouble iz, gdouble amount, gdouble ridges, gdouble *x, gdouble *z);
static gboolean calc_undistorted_coords_out_from_center (gdouble ix, gdouble iz, gdouble amount, gdouble ridges, gdouble *x, gdouble *z);
static gdouble squared(const gdouble x);

///********************************************** gimp's plug-in info *******************************************///
GimpPlugInInfo PLUG_IN_INFO =
{
    NULL,         /// init() optional
    NULL,         /// quit() optional
    query,        /// called when plug-in starts and is changed
    run           /// runs plug-in
};


///******************************************set up default parameters ******************************************///
static plugin_params rbparams =
{
    10,
	5,
	POND_RIPPLES
};


static gint32       img_x1, img_y1, img_x2, img_y2, img_bpp, img_width, img_height, img_has_alpha;
static gdouble		lfCenterX, lfCenterY;
static gdouble		lfXScale, lfYScale;
static gdouble		lfRadius, lfRadMax;

/// c macro that initialises arguments and calls PLUG_IN_INFO
MAIN()


/// query function - responsible for procedure registration and input arguments definition
/// 1. GimpParamDef stores the data so as to speed up loading. only updates if the plug-in is
/// modified. each param has 1-Type; 2-Name; 3-String describing the param
/// 2. gimp_install_procedure declares the procedure name, some description and help strings,
/// menu path where the plug-in should sit, image types handled by the plug-in, and at the
/// end, input and output parameters number, as well as the parameters descriptors.
static void query (void)
{
    static GimpParamDef args[] =
    {
        {
            GIMP_PDB_INT32,
            "run-mode",
            "The run mode { RUN-INTERACTIVE (0), RUN-NONINTERACTIVE (1) }"
        },
        {
            GIMP_PDB_IMAGE,
            "image",
            "Input image"
        },
        {
            GIMP_PDB_DRAWABLE,
            "drawable",
            "Input drawable"
        },
		{
            GIMP_PDB_INT32,
            "amount",
            "Amount of distortion"
        },
		{
            GIMP_PDB_INT32,
            "ridges",
            "Number of ridges"
        },
		{
            GIMP_PDB_INT32,
            "method",
            "Adjustment method used { AROUND_CENTRE (0), OUT_FROM_CENTRE (1), POND_RIPPLES (2) }"
        }
    };

    gimp_install_procedure (PLUG_IN_PROC,
                            PLUGIN_TITLE,
                            "Zigzag image",
                            "Emmanuel Glynn",
                            "Emmanuel Glynn",
                            PLUG_IN_VERSION ,
                            "ZigZag...",
                            "RGB*, GRAY*",
                            GIMP_PLUGIN,
                            G_N_ELEMENTS (args), 0,
                            args, NULL);

    gimp_plugin_menu_register(PLUG_IN_PROC, "<Image>/Filters/Distorts");
}


/// run function - core of the plugin
static void run (const gchar      *name,
                 gint              nparams,
                 const GimpParam  *param,
                 gint             *nreturn_vals,
                 GimpParam       **return_vals)
{
    ///
    static GimpParam    values[1];
    GimpPDBStatusType   status;
    GimpRunMode         run_mode;
    GimpDrawable        *drawable;

    status   = GIMP_PDB_SUCCESS;
    run_mode = (GimpRunMode)param[0].data.d_int32;


    values[0].type          = GIMP_PDB_STATUS;
    values[0].data.d_status = status;

    /// set mandatory output values
    *nreturn_vals = 1;
    *return_vals  = values;

    ///  get the specified drawable
    drawable = gimp_drawable_get (param[2].data.d_drawable);

    img_bpp = gimp_drawable_bpp (drawable->drawable_id);
    img_has_alpha = gimp_drawable_has_alpha (drawable->drawable_id);

	if (!gimp_drawable_mask_intersect (drawable->drawable_id, &img_x1, &img_y1, &img_width, &img_height))
    {
        return;
    }

	img_x2 = img_x1 + img_width;
	img_y2 = img_y1 + img_height;

	lfCenterX = (gdouble)(img_x1 + img_x2 - 1) / 2.0;
	lfCenterY = (gdouble)(img_y1 + img_y2 - 1) / 2.0;

    gdouble xhsize, yhsize;
    xhsize = (gdouble) (img_width - 1) / 2.0;
    yhsize = (gdouble) (img_height - 1) / 2.0;

	if (xhsize > yhsize)
    {
        lfYScale = xhsize / yhsize;
        lfXScale = 1.0;
    }
    else if (yhsize > xhsize)
    {
        lfXScale = yhsize / xhsize;
        lfYScale = 1.0;
    }
    else
    {
        lfXScale = 1.0;
        lfYScale = 1.0;
    }

	lfRadius = (gdouble)MAX(xhsize, yhsize) + 0.5;
	lfRadMax = squared(lfRadius);

    gimp_tile_cache_ntiles (4 * MAX (drawable->width  / gimp_tile_width () + 1 , drawable->height / gimp_tile_height () + 1));

    /// set run modes
    switch( run_mode )
    {
        case GIMP_RUN_INTERACTIVE:
            /// *** get options last values if needed *** ///
            gimp_get_data (PLUG_IN_PROC, &rbparams);

            ///gimp_context_get_foreground (&rbparams.rgb);

            /// *** run the dialog *** ///
            if( !add_method_dialog(drawable) )
                return;
            break;
        case GIMP_RUN_NONINTERACTIVE:
			if ( nparams != 6 )
            {
                status = GIMP_PDB_CALLING_ERROR;
            }
            if ( status == GIMP_PDB_SUCCESS )
            {
                rbparams.amount       = param[3].data.d_int32;
				rbparams.ridges       = param[4].data.d_int32;
				rbparams.method       = param[5].data.d_int32;
            }
            break;
        case GIMP_RUN_WITH_LAST_VALS:
            /// get options last values if needed
            gimp_get_data (PLUG_IN_PROC, &rbparams);
            break;
        default:
            break;
    }
    if ((status == GIMP_PDB_SUCCESS) && (gimp_drawable_is_rgb(drawable->drawable_id) || gimp_drawable_is_gray(drawable->drawable_id)))
    {
        add_plugin_method (drawable);

        if (run_mode != GIMP_RUN_NONINTERACTIVE)
        {
            gimp_displays_flush ();
        }
        if (run_mode == GIMP_RUN_INTERACTIVE)
        {
            gimp_set_data(PLUG_IN_PROC, &rbparams, sizeof(plugin_params ) );
        }
    }
    else if (status == GIMP_PDB_SUCCESS)
    {
        status = GIMP_PDB_EXECUTION_ERROR;
    }

    values[0].data.d_status = status;

    gimp_drawable_detach( drawable );



    return;
}


static void
add_plugin_method(GimpDrawable *drawable)
{
    GimpPixelRgn        des_rgn;
	gpointer			p;
    gint                x, y, ix, iz;
    gint                i;
    guchar             *d, *des;
	gdouble				cx, cy;
	guchar				pixel[4][4];
	guchar              *pixel_elems[4];
	gint				progress, max_progress;
	gdouble 			lfAmount, lfRidges;
	GimpPixelFetcher    *px;


    /// ******************************** initialize PixelRgn to write output data ****************************** ///
    //gimp_pixel_rgn_init (&des_rgn, drawable, 0, 0, img_width, img_height, TRUE, TRUE);
	gimp_pixel_rgn_init (&des_rgn, drawable, img_x1, img_y1, img_width, img_height, TRUE, TRUE);

	progress     = 0;
	max_progress = img_width * img_height;

	gimp_progress_init ("Distorting...");

	px = gimp_pixel_fetcher_new (drawable, FALSE);
    gimp_pixel_fetcher_set_edge_mode (px, GIMP_PIXEL_FETCHER_EDGE_SMEAR);


    if(rbparams.method == AROUND_CENTRE)
    {
		lfAmount = (gdouble)rbparams.amount * G_PI/180.0;
		lfRidges = (G_2_PI * (gdouble)rbparams.ridges) / lfRadius;
        for (p = gimp_pixel_rgns_register (1, &des_rgn), i = 0; p != NULL; p = gimp_pixel_rgns_process (p), i++)
        {
            des = des_rgn.data;
            for (y = des_rgn.y; y < (des_rgn.y + des_rgn.h); ++y)
            {
                d = des;
                for (x = des_rgn.x; x < (des_rgn.x + des_rgn.w); ++x)
                {
                    //progress++;

                    calc_undistorted_coords_around_center ((gdouble)x, (gdouble)y, lfAmount, lfRidges, &cx, &cy);

                    if (cx >= 0.0)
                    {
                        ix = (gint) cx;
                    }
                    else
                    {
                        ix = -((gint) -cx + 1);
                    }

                    if (cy >= 0.0)
                    {
                        iz = (gint) cy;
                    }
                    else
                    {
                        iz = -((gint) -cy + 1);
                    }
                    pixel_elems[0] = pixel[0];
                    pixel_elems[1] = pixel[1];
                    pixel_elems[2] = pixel[2];
                    pixel_elems[3] = pixel[3];

                    gimp_pixel_fetcher_get_pixel (px, ix    , iz    , pixel[0]);
                    gimp_pixel_fetcher_get_pixel (px, ix + 1, iz    , pixel[1]);
                    gimp_pixel_fetcher_get_pixel (px, ix    , iz + 1, pixel[2]);
                    gimp_pixel_fetcher_get_pixel (px, ix + 1, iz + 1, pixel[3]);

                    gimp_bilinear_pixels_8 (d, cx, cy, des_rgn.bpp, img_has_alpha, pixel_elems);

                    d += des_rgn.bpp;
                }
                des += des_rgn.rowstride;
            }
            progress += des_rgn.w * des_rgn.h;
            if (i % 20 == 0)
            {
                gimp_progress_update ((gdouble) progress / (gdouble) max_progress);
            }
        }
    }
    else if(rbparams.method == OUT_FROM_CENTRE)
    {
		lfAmount = (gdouble)rbparams.amount/(G_PI*(gdouble)rbparams.ridges+(gdouble)rbparams.ridges-1);
		lfRidges = (gdouble)rbparams.ridges;
		if(rbparams.ridges != 0)
        {
            for (p = gimp_pixel_rgns_register (1, &des_rgn), i = 0; p != NULL; p = gimp_pixel_rgns_process (p), i++)
            {
                des = des_rgn.data;
                for (y = des_rgn.y; y < (des_rgn.y + des_rgn.h); ++y)
                {
                    d = des;
                    for (x = des_rgn.x; x < (des_rgn.x + des_rgn.w); ++x)
                    {
                        calc_undistorted_coords_out_from_center ((gdouble)x, (gdouble)y, lfAmount, lfRidges, &cx, &cy);

                        if (cx >= 0.0)
                        {
                            ix = (gint) cx;
                        }
                        else
                        {
                            ix = -((gint) -cx + 1);
                        }

                        if (cy >= 0.0)
                        {
                            iz = (gint) cy;
                        }
                        else
                        {
                            iz = -((gint) -cy + 1);
                        }
                        pixel_elems[0] = pixel[0];
                        pixel_elems[1] = pixel[1];
                        pixel_elems[2] = pixel[2];
                        pixel_elems[3] = pixel[3];

                        gimp_pixel_fetcher_get_pixel (px, ix    , iz    , pixel[0]);
                        gimp_pixel_fetcher_get_pixel (px, ix + 1, iz    , pixel[1]);
                        gimp_pixel_fetcher_get_pixel (px, ix    , iz + 1, pixel[2]);
                        gimp_pixel_fetcher_get_pixel (px, ix + 1, iz + 1, pixel[3]);

                        gimp_bilinear_pixels_8 (d, cx, cy, des_rgn.bpp, img_has_alpha, pixel_elems);

                        d += des_rgn.bpp;
                    }
                    des += des_rgn.rowstride;
                }
                progress += des_rgn.w * des_rgn.h;
                if (i % 20 == 0)
                {
                    gimp_progress_update ((gdouble) progress / (gdouble) max_progress);
                }
            }
        }
    }
    else //if(rbparams.method == POND_RIPPLES)
    {
		lfAmount = (gdouble)rbparams.amount;
		lfRidges = (gdouble)rbparams.ridges;
        if(rbparams.ridges != 0)
        {
            for (p = gimp_pixel_rgns_register (1, &des_rgn), i = 0; p != NULL; p = gimp_pixel_rgns_process (p), i++)
            {
                des = des_rgn.data;
                for (y = des_rgn.y; y < (des_rgn.y + des_rgn.h); ++y)
                {
                    d = des;
                    for (x = des_rgn.x; x < (des_rgn.x + des_rgn.w); ++x)
                    {
                        calc_undistorted_coords_pond_ripples ((gdouble)x, (gdouble)y, lfAmount, lfRidges, &cx, &cy);

                        if (cx >= 0.0)
                        {
                            ix = (gint) cx;
                        }
                        else
                        {
                            ix = -((gint) -cx + 1);
                        }

                        if (cy >= 0.0)
                        {
                            iz = (gint) cy;
                        }
                        else
                        {
                            iz = -((gint) -cy + 1);
                        }
                        pixel_elems[0] = pixel[0];
                        pixel_elems[1] = pixel[1];
                        pixel_elems[2] = pixel[2];
                        pixel_elems[3] = pixel[3];

                        gimp_pixel_fetcher_get_pixel (px, ix    , iz    , pixel[0]);
                        gimp_pixel_fetcher_get_pixel (px, ix + 1, iz    , pixel[1]);
                        gimp_pixel_fetcher_get_pixel (px, ix    , iz + 1, pixel[2]);
                        gimp_pixel_fetcher_get_pixel (px, ix + 1, iz + 1, pixel[3]);

                        gimp_bilinear_pixels_8 (d, cx, cy, des_rgn.bpp, img_has_alpha, pixel_elems);

                        d += des_rgn.bpp;
                    }
                    des += des_rgn.rowstride;
                }
                progress += des_rgn.w * des_rgn.h;
                if (i % 20 == 0)
                {
                    gimp_progress_update ((gdouble) progress / (gdouble) max_progress);
                }
            }
        }
    }

    gimp_pixel_fetcher_destroy (px);

    gimp_drawable_flush (drawable);
    gimp_drawable_merge_shadow (drawable->drawable_id, TRUE);
    gimp_drawable_update (drawable->drawable_id, img_x1, img_y1, img_width, img_height);

    return;
}


static gdouble squared(const gdouble x)
{
    return x*x;
}


static gboolean calc_undistorted_coords_pond_ripples (gdouble ix, gdouble iy, gdouble amount, gdouble ridges, gdouble *x, gdouble *y)
{
    gdouble             dx;
    gdouble             dy;
	gboolean			isinside;
	gdouble lfCurrentRadius, lfScale;

	dy = lfYScale*(gdouble)(iy - lfCenterY);
	dx = lfXScale*(gdouble)(ix - lfCenterX);

	lfCurrentRadius = squared(dx) + squared(dy);

	isinside = (lfCurrentRadius <= lfRadMax);

	if (isinside)
	{
		lfCurrentRadius = sqrt(lfCurrentRadius);

		lfScale =   amount * (1.0 - cos(lfCurrentRadius / (lfRadius / ridges) * G_2_PI));

		lfScale *= (lfRadius-lfCurrentRadius)/lfRadius;

		*x = (dx + lfScale) / lfXScale + lfCenterX;
		*y = (dy + lfScale) / lfYScale + lfCenterY;
	}
	else
	{
		*x = ix;
		*y = iy;
	}

	return isinside;
}


static gboolean calc_undistorted_coords_around_center (gdouble ix, gdouble iy, gdouble amount, gdouble ridges, gdouble *x, gdouble *y)
{
	gdouble		dx;
    gdouble		dy;
	gdouble 	lfAngle, lfCurrentRadius, lfCos, lfSin;
	gboolean	isinside;

	dy = lfYScale*(gdouble)(iy - lfCenterY);
	dx = lfXScale*(gdouble)(ix - lfCenterX);

	lfCurrentRadius = squared(dx) + squared(dy);

	isinside = ((lfCurrentRadius <= lfRadMax) );

	if (isinside)
	{
		lfCurrentRadius = sqrt(lfCurrentRadius);

		lfAngle = amount * -(1.0 - sin(1.5 - (lfCurrentRadius * ridges)));
		lfAngle *= (lfRadius-lfCurrentRadius)/lfRadius;

		sincos(lfAngle, &lfSin, &lfCos);

		*x = ((lfCos * dx - lfSin * dy) / lfXScale) + lfCenterX;
		*y = ((lfSin * dx + lfCos * dy) / lfYScale) + lfCenterY;
	}
	else
	{
		*x = ix;
		*y = iy;
	}

	return isinside;
}


static gboolean calc_undistorted_coords_out_from_center (gdouble ix, gdouble iy, gdouble amount, gdouble ridges, gdouble *x, gdouble *y)
{
	gdouble		dx;
    gdouble		dy;
	gdouble 	lfCurrentRadius, lfScale;
	gboolean	isinside;

	dy = lfYScale*(gdouble)(iy - lfCenterY);
	dx = lfXScale*(gdouble)(ix - lfCenterX);

	lfCurrentRadius = squared(dx) + squared(dy);

	isinside = (lfCurrentRadius <= lfRadMax);

	if (isinside)
	{
		lfCurrentRadius = sqrt(lfCurrentRadius);
		lfScale = amount * -(1.0 - cos(lfCurrentRadius * (ridges * G_2_PI)/lfRadius));
		lfScale *= (lfRadius-lfCurrentRadius)/lfRadius;

		if ( lfCurrentRadius != 0.0 )
		{
            lfScale *= (ridges/lfCurrentRadius);
		}

		*x = ix + (dx / lfXScale * lfScale);
		*y = iy + (dy / lfYScale * lfScale);
	}
	else
	{
		*x = ix;
		*y = iy;
	}

	return isinside;
}



static gboolean add_method_dialog(GimpDrawable *drawable)
{
    gboolean    run;

    GtkWidget   *disp_box            = NULL;
    GtkWidget   *dialog              = NULL;
    GtkWidget   *dialog_vbox         = NULL;

    GtkWidget   *scale_3x3table      = NULL;
    GtkObject   *scale               = NULL;

    GtkWidget   *combo               = NULL;


    /// **************************************** initialize plug-in ******************************************** ///
    gimp_ui_init (PLUG_IN_BINARY, FALSE);


    dialog = gimp_dialog_new(PLUGIN_TITLE, PLUG_IN_ROLE,
                             NULL, (GtkDialogFlags) 0,
                             gimp_standard_help_func, PLUG_IN_PROC,
                             GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                             GTK_STOCK_OK,     GTK_RESPONSE_OK,
                             NULL);

    gtk_dialog_set_alternative_button_order (GTK_DIALOG (dialog),
                                             GTK_RESPONSE_OK,
                                             GTK_RESPONSE_CANCEL,
                                             -1);

    gimp_window_set_transient (GTK_WINDOW (dialog));

    /// *************************************** create main horizontal box ************************************* ///
    dialog_vbox = gtk_vbox_new (FALSE, 12);
    /// ******************************************* set border width ******************************************* ///
    gtk_container_set_border_width (GTK_CONTAINER (dialog_vbox), 12);
    /// ************************************ add dialogue box to main box ************************************** ///
    gtk_box_pack_start (GTK_BOX (gtk_dialog_get_content_area (GTK_DIALOG (dialog))), dialog_vbox, TRUE, TRUE, 0);
    /// ***************************************** show main box ************************************************ ///
    gtk_widget_show(dialog_vbox);

    /// ***************************************** prepare preview ********************************************** ///
    disp_box = gimp_zoom_preview_new (drawable);

    /// *************************************** pack preview box into container ******************************** ///
    gtk_box_pack_start (GTK_BOX (dialog_vbox), disp_box, TRUE, TRUE, 0);

    g_signal_connect_swapped (disp_box, "invalidated", G_CALLBACK (add_preview_update), drawable);



    scale_3x3table = gtk_table_new (2, 3, FALSE);
    gtk_table_set_col_spacings (GTK_TABLE (scale_3x3table), 6);
    gtk_table_set_row_spacings (GTK_TABLE (scale_3x3table), 6);
    gtk_box_pack_start (GTK_BOX (dialog_vbox), scale_3x3table, FALSE, FALSE, 0);
    gtk_widget_show (scale_3x3table);

    scale = gimp_scale_entry_new (GTK_TABLE (scale_3x3table),
                                  0,
                                  0,
                                  ("Amount:"),
                                  SCALE_WIDTH,
                                  SPIN_BUTTON_WIDTH,
                                  rbparams.amount,
                                  -100,
                                  100,
                                  1,
                                  8,
                                  0,
                                  TRUE,
                                  0,
                                  0,
                                  NULL,
                                  NULL);
    g_signal_connect (scale, "value_changed", G_CALLBACK (gimp_int_adjustment_update), &rbparams.amount);
    g_signal_connect_swapped (scale, "value-changed", G_CALLBACK (gimp_preview_invalidate), disp_box);

	scale = gimp_scale_entry_new (GTK_TABLE (scale_3x3table),
                                  0,
                                  1,
                                  ("Ridges:"),
                                  SCALE_WIDTH,
                                  SPIN_BUTTON_WIDTH,
                                  rbparams.ridges,
                                  0,
                                  20,
                                  1,
                                  8,
                                  0,
                                  TRUE,
                                  0,
                                  0,
                                  NULL,
                                  NULL);
    g_signal_connect (scale, "value_changed", G_CALLBACK (gimp_int_adjustment_update), &rbparams.ridges);
    g_signal_connect_swapped (scale, "value-changed", G_CALLBACK (gimp_preview_invalidate), disp_box);

    combo = gimp_int_combo_box_new (("Around Center"), AROUND_CENTRE,
                                    ("Out From Center"), OUT_FROM_CENTRE,
                                    ("Pond Ripples"), POND_RIPPLES,
                                    NULL);

    gimp_int_combo_box_connect (GIMP_INT_COMBO_BOX (combo),
                                rbparams.method,
                                G_CALLBACK (gimp_int_combo_box_get_active),
                                &rbparams.method);

    gimp_table_attach_aligned (GTK_TABLE (scale_3x3table), 0, 2, ("Style:"), 0.0, 0.5, combo, 2, FALSE);
    g_signal_connect_swapped (combo, "changed", G_CALLBACK (gimp_preview_invalidate), disp_box);

    /// ***************************************** show dialog box ********************************************** ///
    gtk_widget_show(disp_box);

    /// ******************************************** run dialog ************************************************ ///
    run = (gimp_dialog_run(GIMP_DIALOG(dialog)) == GTK_RESPONSE_OK);
    /// ********************************************* clean up ************************************************* ///
    gtk_widget_destroy(dialog);

    return run;
}


static void add_preview_update(GimpDrawable *drawable, GimpPreview *preview_box)
{
	GimpPixelFetcher    *p;
    guchar   			*des, *d;
    gint      			width, height;
    gint      			bpp;
    gint      			x, y, px, py;
    gdouble             ix, iy;
	guchar				pixel0[4], pixel1[4], pixel2[4], pixel3[4];
	guchar				*pixel_elems[4];
	gdouble				lfAmount, lfRidges;

	pixel_elems[0] = pixel0;
	pixel_elems[1] = pixel1;
	pixel_elems[2] = pixel2;
	pixel_elems[3] = pixel3;

    p = gimp_pixel_fetcher_new (drawable, FALSE);
    gimp_pixel_fetcher_set_edge_mode (p, GIMP_PIXEL_FETCHER_EDGE_SMEAR);

    des  = gimp_zoom_preview_get_source (GIMP_ZOOM_PREVIEW (preview_box), &width, &height, &bpp);


	d = des;
	if(rbparams.method == AROUND_CENTRE)
    {
		lfAmount = (gdouble)rbparams.amount * G_PI/180.0;
		lfRidges = (G_2_PI * (gdouble)rbparams.ridges) / lfRadius;
        for (y = 0; y < height; ++y)
        {
            for (x = 0; x < width; ++x)
            {
                gimp_preview_untransform (preview_box, x, y, &px, &py);
                if(calc_undistorted_coords_around_center ((gdouble)px, (gdouble)py, lfAmount, lfRidges, &ix, &iy))
                {
                    gimp_pixel_fetcher_get_pixel (p, ix    , iy    , pixel0);
                    gimp_pixel_fetcher_get_pixel (p, ix + 1, iy    , pixel1);
                    gimp_pixel_fetcher_get_pixel (p, ix    , iy + 1, pixel2);
                    gimp_pixel_fetcher_get_pixel (p, ix + 1, iy + 1, pixel3);


                    gimp_bilinear_pixels_8 (d, ix, iy, bpp, img_has_alpha, pixel_elems);
                }

                d += bpp;
            }
        }
    }
    else if(rbparams.method == OUT_FROM_CENTRE)
    {
		lfAmount = (gdouble)rbparams.amount/(G_PI*(gdouble)rbparams.ridges+(gdouble)rbparams.ridges-1);
		lfRidges = (gdouble)rbparams.ridges;
		if(rbparams.ridges != 0)
        {
            for (y = 0; y < height; ++y)
            {
                for (x = 0; x < width; ++x)
                {
                    gimp_preview_untransform (preview_box, x, y, &px, &py);
                    if(calc_undistorted_coords_out_from_center ((gdouble)px, (gdouble)py, lfAmount, lfRidges, &ix, &iy))
                    {
                        gimp_pixel_fetcher_get_pixel (p, ix    , iy    , pixel0);
                        gimp_pixel_fetcher_get_pixel (p, ix + 1, iy    , pixel1);
                        gimp_pixel_fetcher_get_pixel (p, ix    , iy + 1, pixel2);
                        gimp_pixel_fetcher_get_pixel (p, ix + 1, iy + 1, pixel3);


                        gimp_bilinear_pixels_8 (d, ix, iy, bpp, img_has_alpha, pixel_elems);
                    }

                    d += bpp;
                }
            }
        }
    }
    else //if(rbparams.method == POND_RIPPLES)
    {
		lfAmount = (gdouble)rbparams.amount;
		lfRidges = (gdouble)rbparams.ridges;
        if(rbparams.ridges != 0)
        {
            for (y = 0; y < height; ++y)
            {
                for (x = 0; x < width; ++x)
                {
                    gimp_preview_untransform (preview_box, x, y, &px, &py);
                    if(calc_undistorted_coords_pond_ripples ((gdouble)px, (gdouble)py, lfAmount, lfRidges, &ix, &iy))
                    {
                        gimp_pixel_fetcher_get_pixel (p, ix    , iy    , pixel0);
                        gimp_pixel_fetcher_get_pixel (p, ix + 1, iy    , pixel1);
                        gimp_pixel_fetcher_get_pixel (p, ix    , iy + 1, pixel2);
                        gimp_pixel_fetcher_get_pixel (p, ix + 1, iy + 1, pixel3);


                        gimp_bilinear_pixels_8 (d, ix, iy, bpp, img_has_alpha, pixel_elems);
                    }

                    d += bpp;
                }
            }
        }
    }

    gimp_pixel_fetcher_destroy (p);

    gimp_preview_draw_buffer (preview_box, des, width * bpp);

    g_free (des);
}


